package de.mrbesen.telegram.commands;

import de.mrbesen.telegram.AsyncHandler;
import de.mrbesen.telegram.MessageBuilder;
import de.mrbesen.telegram.TelegramAPI;
import de.mrbesen.telegram.event.events.UserCallbackEvent;
import de.mrbesen.telegram.event.events.UserSendMessageEvent;
import de.mrbesen.telegram.objects.TInlineKeyboardMarkup;
import de.mrbesen.telegram.objects.TMessage;
import de.mrbesen.telegram.objects.TUser;

import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

public class FeedbackCommand implements JSONCommandHandler {

	protected final TInlineKeyboardMarkup cancelMarkup = new TInlineKeyboardMarkup(1);
	protected final Map<Long, TMessage> awaitingFeedback = new TreeMap<>(); //maps users chat id -> message that said "please send feedback"  TODO: make persistent?
	protected final Map<Long, Long[]> awnser = new TreeMap<>(); // chatid -> [chatid, messageid] to awnser to; next message from key will be fwd to value[0] and reply to value[1]
	protected final TelegramAPI api;
	protected final List<Long> admins;
	protected final static String CANCELFEEDBACK = "cancelfeedback", REPLYFEEDBACK = "replyFeedback ";


	public FeedbackCommand(TelegramAPI api, List<Long> admins) {
		this.api = api;
		this.admins = (admins == null ? new LinkedList<>() : admins);

		//build markup
		cancelMarkup.addCallbackButton("❌ cancel", CANCELFEEDBACK, 1);
	}

	@Override
	public boolean onCommand(TUser sender, String cmd, String[] args, TMessage json) {
		if(json.getChatID() < 0) return false; //do not handle group chats

		if(cmd.equalsIgnoreCase("feedback")) {
			removeMsg(sender.getID(), null);
			awaitingFeedback.put(sender.getID(), null);
			MessageBuilder mb = new MessageBuilder();
			mb.setAsync().setReciver(sender).setMarkup(cancelMarkup).setText("Your next Message will be forwarded to the admins.")
					.setCallback(new AsyncHandler.Callback<TMessage, Object>() {
						@Override
						public Object call(TMessage msg) {
							awaitingFeedback.put(sender.getID(), msg);
							return null;
						}
					});
			api.sendMessage(mb.build());
			return true;
		}
		return false;
	}

	public boolean onMsg(UserSendMessageEvent e) {
		TUser user = e.getUser();
		TMessage msg = e.getMessage();
		System.out.println("msg: " + msg.getText());
		if (awaitingFeedback.containsKey(user.getID())) {
			removeMsg(user.getID(), e.getMessage());
			if(isFeedback(msg.getText())) {
				//fwd to admins
				admins.forEach(a -> TMessage.forwardAsync(api, a, msg));

				MessageBuilder mb = createInfo(user, msg.getMessageID());
				admins.forEach(a -> api.sendMessage(mb.setReciver(a).build()));
			} else {
				System.out.println("feedback omitted: " + msg.getText());
			}
			return true;
		} else {
			Long[] rplymsg = awnser.remove(user.getID());
			if(rplymsg != null) {
				if(admins.contains(user.getID())) {
					System.out.println("is admin!");
					//admin to user -> copy text send as new msg from bot to disguise admin
					TInlineKeyboardMarkup mu = new TInlineKeyboardMarkup(1);
					mu.addCallbackButton("↩ reply", buildRplyCallback(user.getID(), msg.getMessageID()), 1);

					MessageBuilder mb = new MessageBuilder().setAsync().setReciver(rplymsg[0])
							.setReplyTo(Math.toIntExact(rplymsg[1])).setText(msg.getText()).setMarkup(mu);
					api.sendMessage(mb.build());

					awnser.put(rplymsg[0], new Long[]{user.getID(), Long.valueOf(msg.getMessageID())});
				} else {
					System.out.println("is user!");
					//user to admin -> fwd message
					TMessage.forwardAsync(api, rplymsg[0], msg);

					//send info
					MessageBuilder mb = createInfo(user, msg.getMessageID());
					api.sendMessage(mb.setReciver(rplymsg[0]).setReplyTo(Math.toIntExact(rplymsg[1])).build());
				}
				System.out.println("Feedback reply from " + user.getID() + " to " + rplymsg[0] + ":" + rplymsg[1] + " send.");
				return true;
			}
		}
		return false;
	}

	protected MessageBuilder createInfo(TUser u, int msgid) {
		String info = "Feedbackinfo:\n" + "userid: " + u.getID() + "\nusername: " + u.getName() + "\nfullname: " + u.getFirstName() + ' ' + u.getLastName();
		TInlineKeyboardMarkup markup = new TInlineKeyboardMarkup(1);
		markup.addUrlButton("chat with user", "https://t.me/" + u.getName(), 1);
		markup.addCallbackButton("↩ reply", buildRplyCallback(u.getID(), msgid), 1);
		return new MessageBuilder().setAsync().setText(info).setMarkup(markup);
	}

	public boolean onCallback(UserCallbackEvent e) {
		if(e.getData().equals(CANCELFEEDBACK)) {
			removeMsg(e.getUser().getID(), e.getMsg());
			api.answerCallbackQuery(e.getID(), "Cancelled", true);
			return true;
		} else if(e.getData().startsWith(REPLYFEEDBACK)) {
			String rest = e.getData().substring(REPLYFEEDBACK.length());
			long chatid = 0;
			long msgid = -1;
			try {
				String[] spl = rest.split(" ",2);
				long adminid = Long.parseLong(spl[0]);
				msgid = Integer.parseInt(spl[1]);
				//convert admin id to chat id
				if(adminid <= 0) {
					chatid = admins.get((int) -adminid);
				} else
					chatid = adminid;
			} catch (NumberFormatException ex) {
				api.answerCallbackQuery(e.getID(), "Error. Sorry :/", true);
				ex.printStackTrace();
			}
			if(msgid > -1) {
				awnser.put(e.getUser().getID(), new Long[]{chatid, msgid});
				api.answerCallbackQuery(e.getID(), "Write a Message", true);
			}
			return true;
		}
		return false;
	}

	/**
	 * looks up msg from awaiting feedback map, if not present try to use fallback_msg
	 * is Async
	 * removes user from awaitingFeedback
	 *
	 */
	protected void removeMsg(long chat, TMessage fallback_msg) {
		TMessage msg = awaitingFeedback.remove(chat);
		if(msg == null)	msg = fallback_msg;

		if(msg != null) TMessage.delete(api, msg.getChatID(), msg.getMessageID(), true);
	}

	/**
	 * check if a message is allowed as feedback
	 * @param msg
	 * @return
	 */
	protected boolean isFeedback(String msg) {
		msg = msg.trim();
		return (!msg.startsWith("/")) && (msg.length() > 5) && (!msg.startsWith("http"));
	}

	protected int getAdminID(long adminchat) {
		for(int i = 0; i < admins.size(); i++) {
			if(admins.get(i) == adminchat)
				return i;
		}
		return -1;
	}

	protected String buildRplyCallback(long user, int msgid) {
		int adminid = getAdminID(user);
		if(adminid >= 0) user = -adminid;
		return REPLYFEEDBACK + user + " " + msgid;
	}
}
